r"""
This module is to simulate arbitrary system heterogeneity that may occur in practice.
We conclude four types of system heterogeneity from existing works.
System Heterogeneity Description:
    1. **Availability**: the devices will be either available or unavailable at each moment, where only the
                    available devices can be selected to participate in training.

    2. **Responsiveness**: the responsiveness describes the length of the period from the server broadcasting the
                    gloabl model to the server receiving the locally trained model from a particular client.

    3. **Completeness**: since the server cannot fully control the behavior of devices,it's possible for devices to
                    upload imcomplete model updates (i.e. only training for a few steps).

    4. **Connectivity**: the clients who promise to complete training may suffer accidients so that the server may lose
                    connections with these client who will never return the currently trained local_movielens_recommendation model.

We build up a client state machine to simulate the four types of system heterogeneity, and provide high-level
APIs to allow customized system heterogeneity simulation.

**Example**: How to customize the system heterogeneity:
```python
>>> class MySimulator(flgo.simulator.base.BasicSimulator):
...     def update_client_availability(self):
...         # update the variable 'prob_available' and 'prob_unavailable' for all the clients
...         self.set_variable(self.all_clients, 'prob_available', [0.9 for _ in self.all_clients])
...         self.set_variable(self.all_clients, 'prob_unavailable', [0.1 for _ in self.all_clients])
...
...     def update_client_connectivity(self, client_ids):
...         # update the variable 'prob_drop' for clients in client_ids
...         self.set_variable(client_ids, 'prob_drop', [0.1 for _ in client_ids])
...
...     def update_client_responsiveness(self, client_ids, *args, **kwargs):
...         # update the variable 'latency' for clients in client_ids
...         self.set_variable(client_ids, 'latency', [np.random.randint(5,100) for _ in client_ids])
...
...     def update_client_completeness(self, client_ids, *args, **kwargs):
...         # update the variable 'working_amount' for clients in client_ids
...         self.set_variable(client_ids, 'working_amount',  [max(int(self.clients[cid].num_steps*np.random.rand()), 1) for cid in client_ids])
>>> r = flgo.init(task, algorithm=fedavg, Simulator=MySimulator)
>>> # The runner r will be runned under the customized system heterogeneity, where the clients' states will be flushed by
>>> # MySimulator.update_client_xxx at each moment of the virtual clock or particular events happen (i.e. a client was selected)
```

We also provide some preset Simulator like flgo.simulator.DefaultSimulator and flgo.simulator.
"""
import os

from flgo.simulator.default_simulator import Simulator as DefaultSimulator
from flgo.simulator.phone_simulator import Simulator as PhoneSimulator
from flgo.simulator.base import BasicSimulator
import numpy as np
import random
import matplotlib.pyplot as plt

class ResponsivenessExampleSimulator(BasicSimulator):
    def initialize(self):
        self.client_time_response = {cid: np.random.randint(5, 1000) for cid in self.clients}
        self.set_variable(list(self.clients.keys()), 'latency', list(self.client_time_response.values()))

    def update_client_responsiveness(self, client_ids):
        latency = [self.client_time_response[cid] for cid in client_ids]
        self.set_variable(client_ids, 'latency', latency)

class CompletenessExampleSimulator(BasicSimulator):
    def update_client_completeness(self, client_ids):
        if not hasattr(self, '_my_working_amount'):
            rs = self.random_module.normal(1.0, 1.0, len(self.clients))
            rs = rs.clip(0.01, 2)
            self._my_working_amount = {cid:max(int(r*self.clients[cid].num_steps),1) for  cid,r in zip(self.clients, rs)}
        working_amount = [self._my_working_amount[cid] for cid in client_ids]
        self.set_variable(client_ids, 'working_amount', working_amount)

class AvailabilityExampleSimulator(BasicSimulator):
    def update_client_availability(self):
        if self.gv.clock.current_time==0:
            self.set_variable(self.all_clients, 'prob_available', [1.0 for _ in self.clients])
            self.set_variable(self.all_clients, 'prob_unavailable', [0.0 for _ in self.clients])
            return
        pa = [0.1 for _ in self.clients]
        pua = [0.1 for _ in self.clients]
        self.set_variable(self.all_clients, 'prob_available', pa)
        self.set_variable(self.all_clients, 'prob_unavailable', pua)

class ConnectivityExampleSimulator(BasicSimulator):
    def initialize(self):
        drop_probs = self.random_module.uniform(0.,0.05, len(self.clients)).tolist()
        self.client_drop_prob = {cid: dp for cid,dp in zip(self.clients, drop_probs)}

    def update_client_connectivity(self, client_ids):
        self.set_variable(client_ids, 'prob_drop', [self.client_drop_prob[cid] for cid in client_ids])

class ExampleSimulator(BasicSimulator):
    def initialize(self):
        drop_probs = self.random_module.uniform(0.,0.05, len(self.clients)).tolist()
        self.client_drop_prob = {cid: dp for cid,dp in zip(self.clients, drop_probs)}
        self.client_time_response = {cid: np.random.randint(5, 1000) for cid in self.clients}
        self.set_variable(list(self.clients.keys()), 'latency', list(self.client_time_response.values()))
        self.server.tolerance_for_latency = 999999

    def update_client_connectivity(self, client_ids):
        self.set_variable(client_ids, 'prob_drop', [self.client_drop_prob[cid] for cid in client_ids])

    def update_client_responsiveness(self, client_ids):
        latency = [self.client_time_response[cid] for cid in client_ids]
        self.set_variable(client_ids, 'latency', latency)

    def update_client_availability(self):
        if self.gv.clock.current_time==0:
            self.set_variable(self.all_clients, 'prob_available', [1.0 for _ in self.clients])
            self.set_variable(self.all_clients, 'prob_unavailable', [0.0 for _ in self.clients])
            return
        pa = [0.1 for _ in self.clients]
        pua = [0.1 for _ in self.clients]
        self.set_variable(self.all_clients, 'prob_available', pa)
        self.set_variable(self.all_clients, 'prob_unavailable', pua)

    def update_client_completeness(self, client_ids):
        if not hasattr(self, '_my_working_amount'):
            rs = self.random_module.normal(1.0, 1.0, len(self.clients))
            rs = rs.clip(0.01, 2)
            self._my_working_amount = {cid:max(int(r*self.clients[cid].num_steps),1) for  cid,r in zip(self.clients, rs)}
        working_amount = [self._my_working_amount[cid] for cid in client_ids]
        self.set_variable(client_ids, 'working_amount', working_amount)

def visualize_availability(data, sort=True, title='', show=True, save=False):
    """
    Visualize availability matrix

    Args:
        data (numpy.ndarray): a 2d-array where each row corresponds to round-wise availability of all the clients and each column refers to the availability of one client across rounds.
    """
    if not isinstance(data, np.ndarray):data = np.array(data)
    data = data.T
    if sort:
        row_sums = np.sum(data, axis=1)
        sorted_indices = np.argsort(row_sums)
        data = data[sorted_indices]
    fig, ax1 = plt.subplots()
    ax1.imshow(data, cmap='Greens')
    bars = np.sum(data, axis=1)
    ax1.barh(list(range(len(bars))), bars, alpha=0.25)
    ax1.set_xlabel('Round')
    ax1.set_ylabel('Client ID')
    data2 = np.sum(data, axis=0)
    ax1.plot(list(range(len(data2))), data2, 'r', label='num')
    tit = 'Client Availability'
    if title!='': tit = tit + '-' + title
    ax1.invert_yaxis()
    plt.title(tit)
    plt.grid(False)
    plt.tight_layout()
    if save: plt.savefig(f'{tit}.png', dpi=300, bbox_inches='tight')
    if show: plt.show()
    return tit

def visualize_capacity(data, sort=True, title='', show=True, save=False):
    if not isinstance(data, np.ndarray):data = np.array(data)
    data = data.T
    if sort:
        row_sums = np.sum(data, axis=1)
        sorted_indices = np.argsort(row_sums)
        data = data[sorted_indices]
    mean_cap = np.mean(data, axis=1)
    min_cap = np.min(data, axis=1)
    max_cap = np.max(data, axis=1)
    plt.figure()
    plt.bar(list(range(len(mean_cap))), mean_cap, color='gray')
    plt.plot(list(range(len(mean_cap))), mean_cap, linewidth=2, color='black')
    plt.scatter(list(range(len(min_cap))), min_cap, s=6, marker='o', color='green')
    plt.scatter(list(range(len(min_cap))), max_cap, s=6, marker='o', color='red')
    plt.xlabel("Client ID")
    plt.ylabel("Capacity Ratio")
    tit = "Device Capacity"
    if title!='':tit=tit + '-' + title
    plt.title(tit)
    plt.tight_layout()
    if save: plt.savefig(f'{tit}.png', dpi=300, bbox_inches='tight')
    if show: plt.show()
    return tit

def visualize_latency(data, sort=True, title='', show=True, save=False):
    if not isinstance(data, np.ndarray):data = np.array(data)
    if sort:
        h = data.T
        rows = np.mean(h, axis=1)
        sorted_indices = np.argsort(rows)
        h = h[sorted_indices]
        data = h.T
    means = np.mean(data, axis=0)
    fig, ax1 = plt.subplots()
    ax2 = ax1.twinx()
    ax2.barh(np.arange(len(means)), means, color='red', alpha=0.2)
    ax2.set_ylabel('Clients', color='red')
    ax2.tick_params(axis='y', labelcolor='red')
    # ax2.scatter(full_max, len(means), color='red', s=100, marker='*', label='Full Par.')
    rounds = data.shape[0]
    colors = ['red', 'pink', 'yellow', 'green', 'purple']
    ps = [1.0, 0.5,0.2, 0.1, 0.01, ]
    all_d = []
    all_n = []
    for p,c in zip(ps, colors):
        ds = []
        for k in range(20):
            d = np.array([np.random.choice(data[i], size=max(int(data.shape[1] * p), 1), replace=False).tolist() for i in range(rounds)]).max(axis=1).mean()
            ds.append(d)
        d = np.array(ds).mean()
        all_d.append(d)
        all_n.append(len(means)*p)
        # ax2.scatter(d, len(means)*p, color=c, s=100, marker='*', label=f'{int(p*100)}% Par.')
    ax2.plot(all_d, all_n, '--', color='black')
    for d,n,c,p in zip(all_d, all_n, colors,ps):
        ax2.scatter(d, n, color=c, s=100, marker='*', label=f'{int(p*100)}% Par.')
    hist_values, bins, patches = ax1.hist(means, bins='auto',edgecolor='none', color='skyblue')
    ax1.set_xlabel('Latency (virtual time unit)')
    ax1.set_ylabel('Num of Client', color='blue')
    ax1.tick_params(axis='y', labelcolor='blue')
    bin_centers = 0.5 * (bins[1:] + bins[:-1])  # 计算每个bin的中心
    if bin_centers.size>2:
        from scipy.interpolate import make_interp_spline
        x_smooth = np.linspace(bin_centers.min(), bin_centers.max(), 300)  # 更高分辨率的 x 数据
        spline = make_interp_spline(bin_centers, hist_values, k=2)  # 使用三次样条插值
        y_smooth = spline(x_smooth)
        ax1.plot(x_smooth, y_smooth, color='blue', linestyle='-', linewidth=2)
        ax1.scatter(bin_centers, hist_values, color='blue', marker='o')
    else:
        ax1.scatter(bin_centers, hist_values, color='blue', marker='o')
    tit = "Responsiveness"
    if title!='':tit=tit + '-' + title
    plt.title(tit)
    plt.legend(loc='upper left')
    plt.tight_layout()
    if save: plt.savefig(f'{tit}.png', dpi=300, bbox_inches='tight')
    if show: plt.show()
    return tit

def visualize_completeness(data, sort=True, title="", show=True, save=False):
    if not isinstance(data, np.ndarray):data = np.array(data).T
    if sort:
        rows = np.mean(data, axis=1)
        sorted_indices = np.argsort(rows)
        data = data[sorted_indices]
    fig, ax = plt.subplots()
    cax = ax.imshow(data, cmap='Reds', interpolation='nearest')
    fig.colorbar(cax, ax=ax, label='Completeness Degree',  fraction=0.046)
    tit = "Training Completeness"
    if title!='':tit=tit + '-' + title
    plt.title(tit)
    plt.gca().invert_yaxis()
    plt.ylabel('Client ID')
    plt.xlabel('Rounds')
    plt.tight_layout()
    # plt.axis('equal')
    # plt.ylim([0, data.shape[0]])
    ax.set_aspect('auto', adjustable='datalim')
    ax.set_ylim([0, data.shape[0]])
    if save: plt.savefig(f'{tit}.png', dpi=300, bbox_inches='tight')
    if show: plt.show()
    return tit

def visualize_stability(data, sort=True, title="", show=True, save=False):
    if not isinstance(data, np.ndarray): data = np.array(data).T
    if sort:
        rows = np.sum(data[:,0,:], axis=1)
        sorted_indices = np.argsort(rows)
        data = data[sorted_indices]
    dropped = data[:,0,:]
    probs = data[:,1,:]
    fig, ax1 = plt.subplots()
    ax1.set_ylabel('P(Dropping)', color='blue')
    ax1.set_xlabel('Client ID')
    ax1.boxplot(probs.T)

    ax2 = ax1.twinx()
    ax2.set_ylabel('Actual Dropping Times', color='orange')
    dropped = dropped.sum(axis=1)
    ax2.bar(list(range(len(dropped))), dropped, color='orange', alpha=0.2)
    tit = "Connection Stability"
    if title!='':tit=tit + '-' + title
    plt.title(tit)
    plt.xticks([])
    plt.tight_layout()
    if save: plt.savefig(f'{tit}.png', dpi=300, bbox_inches='tight')
    if show: plt.show()
    return tit

def visualize_simulator(runner, sort=False, together=True, save=True, select=[], title=""):
    """
    Visualize a simulator.

    Args:
        runner (Any): the runner generated by flgo.init
        sort (bool): reasonably sorting clients in each system heterogeneity figure
        together (bool): plot all the figures together in one figure
        save (bool): save the figure to disk
        select (list): select which system heterogeneity to be visualized. Default is selecting all. [0,1,2,3,4] corresponds to [availability, responsiveness, completeness, connectivity,capacity] respectively.
    """
    runner.proportion = 1.0
    runner.sample_option = 'full'
    runner.eval_interval = -1
    completeness = []
    latencies = []
    avails = []
    drops = []
    caps = []
    if select==[]: select=list(range(5))
    def pack(self, client_id, mtype=0):
        return {}

    def iterate(self):
        caps.append([c._capacity for c in self.clients])
        avails.append([int(c.is_idle()) for c in self.clients])
        self.selected_clients = list(range(self.num_clients))
        res = self.communicate(self.selected_clients)
        dropped = [1. if cid in self._dropped_selected_clients else 0. for cid in self.selected_clients]
        pdrop = self.gv.simulator.get_variable(self.gv.simulator.idx2id(self.selected_clients), 'prob_drop')
        drops.append([dropped, pdrop])
        completeness.append([1.0*c._working_amount/c.num_steps for c in self.clients])
        latencies.append([c._latency for c in self.clients])
        return

    def reply(self, svr_pkg):
        return {}

    runner.__class__.pack = pack
    runner.__class__.iterate = iterate
    for c in runner.clients:
        c.__class__.reply = reply
        c.actions = {0: c.reply}
    runner.run()
    n = f"{runner.gv.simulator.__class__.__name__}-R{runner.num_rounds}"
    if together:
        ts = []
        if 0 in select: ts.append(visualize_availability(avails, sort, show=False, save=True))
        if 1 in select: ts.append(visualize_latency(latencies, sort, show=False, save=True))
        if 2 in select: ts.append(visualize_completeness(completeness, sort, show=False, save=True))
        if 3 in select: ts.append(visualize_stability(drops, sort, show=False, save=True))
        if 4 in select: ts.append(visualize_capacity(caps, sort, show=False, save=True))
        plt.close('all')
        image_files = [f"{t}.png" for t in ts]
        import matplotlib.image as mpimg
        fig, axs = plt.subplots(1, len(select), figsize=(4*len(select), 4))
        for ax, image_file in zip(axs, image_files):
            img = mpimg.imread(image_file)  # 读取图像
            ax.imshow(img)  # 显示图像
            ax.axis('off')  # 关闭坐标轴
        plt.suptitle(title if title!='' else n)
        plt.tight_layout()
        if save: plt.savefig(f'{n}.png', dpi=300, bbox_inches='tight')
        plt.show()
        for img in image_files: os.remove(img)
    else:
        n = title if title!='' else n
        if 0 in select: visualize_availability(avails, sort, n, save=save)
        if 1 in select: visualize_latency(latencies, sort, n, save=save)
        if 2 in select: visualize_completeness(completeness, sort, n, save=save)
        if 3 in select: visualize_stability(drops, sort, n, save=save)
        if 4 in select: visualize_capacity(caps, sort, n, save=save)










"""resnet in pytorch



[1] Kaiming He, Xiangyu Zhang, Shaoqing Ren, Jian Sun.

    Deep Residual Learning for Image Recognition
    https://arxiv.org/abs/1512.03385v1
"""

import torch
import torch.nn as nn
import torch
import torch.nn as nn
import flgo.utils.fmodule as fuf
from torch.utils.data import Dataset
import torchvision.transforms as transforms
import torchvision
class Model(fuf.FModule):
    def __init__(self):
        super().__init__()
        self.model = resnet18()

    def forward(self, *args, **kwargs):
        return self.model(*args, **kwargs)

class AugmentDataset(Dataset):
    def __init__(self, dataset):
        self.dataset = dataset
        self.transform = torchvision.transforms.Compose([transforms.RandomCrop(size=(32, 32), padding=4), transforms.RandomHorizontalFlip(0.5)])

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, item):
        img, label = self.dataset[item]
        return self.transform(img), label

def init_dataset(object):
    if 'Client' in object.get_classname():
        object.train_data = AugmentDataset(object.train_data)

def init_local_module(object):
    pass

def init_global_module(object):
    if 'Server' in object.__class__.__name__:
        object.model = Model().to(object.device)
class BasicBlock(nn.Module):
    """Basic Block for resnet 18 and resnet 34

    """

    #BasicBlock and BottleNeck block
    #have different output size
    #we use class attribute expansion
    #to distinct
    expansion = 1

    def __init__(self, in_channels, out_channels, stride=1):
        super().__init__()

        #residual function
        self.residual_function = nn.Sequential(
            nn.Conv2d(in_channels, out_channels, kernel_size=3, stride=stride, padding=1, bias=False),
            nn.GroupNorm(2, out_channels),
            nn.ReLU(inplace=True),
            nn.Conv2d(out_channels, out_channels * BasicBlock.expansion, kernel_size=3, padding=1, bias=False),
            nn.GroupNorm(2,out_channels * BasicBlock.expansion)
        )

        #shortcut
        self.shortcut = nn.Sequential()

        #the shortcut output dimension is not the same with residual function
        #use 1*1 convolution to match the dimension
        if stride != 1 or in_channels != BasicBlock.expansion * out_channels:
            self.shortcut = nn.Sequential(
                nn.Conv2d(in_channels, out_channels * BasicBlock.expansion, kernel_size=1, stride=stride, bias=False),
                nn.GroupNorm(2,out_channels * BasicBlock.expansion)
            )

    def forward(self, x):
        return nn.ReLU(inplace=True)(self.residual_function(x) + self.shortcut(x))

class BottleNeck(nn.Module):
    """Residual block for resnet over 50 layers

    """
    expansion = 4
    def __init__(self, in_channels, out_channels, stride=1):
        super().__init__()
        self.residual_function = nn.Sequential(
            nn.Conv2d(in_channels, out_channels, kernel_size=1, bias=False),
            nn.GroupNorm(2, out_channels),
            nn.ReLU(inplace=True),
            nn.Conv2d(out_channels, out_channels, stride=stride, kernel_size=3, padding=1, bias=False),
            nn.GroupNorm(2, out_channels),
            nn.ReLU(inplace=True),
            nn.Conv2d(out_channels, out_channels * BottleNeck.expansion, kernel_size=1, bias=False),
            nn.GroupNorm(2,out_channels * BottleNeck.expansion),
        )

        self.shortcut = nn.Sequential()

        if stride != 1 or in_channels != out_channels * BottleNeck.expansion:
            self.shortcut = nn.Sequential(
                nn.Conv2d(in_channels, out_channels * BottleNeck.expansion, stride=stride, kernel_size=1, bias=False),
                nn.GroupNorm(2,out_channels * BottleNeck.expansion)
            )

    def forward(self, x):
        return nn.ReLU(inplace=True)(self.residual_function(x) + self.shortcut(x))

class ResNet(nn.Module):

    def __init__(self, block, num_block, num_classes=10):
        super().__init__()

        self.in_channels = 64

        self.conv1 = nn.Sequential(
            nn.Conv2d(3, 64, kernel_size=3, padding=1, bias=False),
            nn.GroupNorm(2,64),
            nn.ReLU(inplace=True))
        #we use a different inputsize than the original paper
        #so conv2_x's stride is 1
        self.conv2_x = self._make_layer(block, 64, num_block[0], 1)
        self.conv3_x = self._make_layer(block, 128, num_block[1], 2)
        self.conv4_x = self._make_layer(block, 256, num_block[2], 2)
        self.conv5_x = self._make_layer(block, 512, num_block[3], 2)
        self.avg_pool = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = nn.Linear(512 * block.expansion, num_classes)

    def _make_layer(self, block, out_channels, num_blocks, stride):
        """make resnet layers(by layer i didnt mean this 'layer' was the
        same as a neuron netowork layer, ex. conv layer), one layer may
        contain more than one residual block

        Args:
            block: block type, basic block or bottle neck block
            out_channels: output depth channel number of this layer
            num_blocks: how many blocks per layer
            stride: the stride of the first block of this layer

        Return:
            return a resnet layer
        """

        # we have num_block blocks per layer, the first block
        # could be 1 or 2, other blocks would always be 1
        strides = [stride] + [1] * (num_blocks - 1)
        layers = []
        for stride in strides:
            layers.append(block(self.in_channels, out_channels, stride))
            self.in_channels = out_channels * block.expansion

        return nn.Sequential(*layers)

    def forward(self, x):
        output = self.conv1(x)
        output = self.conv2_x(output)
        output = self.conv3_x(output)
        output = self.conv4_x(output)
        output = self.conv5_x(output)
        output = self.avg_pool(output)
        output = output.view(output.size(0), -1)
        output = self.fc(output)

        return output

def resnet18():
    """ return a ResNet 18 object
    """
    return ResNet(BasicBlock, [2, 2, 2, 2])

def resnet34():
    """ return a ResNet 34 object
    """
    return ResNet(BasicBlock, [3, 4, 6, 3])

def resnet50():
    """ return a ResNet 50 object
    """
    return ResNet(BottleNeck, [3, 4, 6, 3])

def resnet101():
    """ return a ResNet 101 object
    """
    return ResNet(BottleNeck, [3, 4, 23, 3])

def resnet152():
    """ return a ResNet 152 object
    """
    return ResNet(BottleNeck, [3, 8, 36, 3])





from .model import default_model
import flgo.benchmark.toolkits.visualization as v
import flgo.benchmark.toolkits.partition

default_model = default_model
default_partitioner = flgo.benchmark.toolkits.partition.BasicHierPartitioner
default_partition_para = {}
visualize = v.visualize_hier_by_class


import copy
import os
import torch.nn as nn
import numpy as np
import torch.nn.functional as F
from tqdm import trange
import flgo.algorithm.fedbase as fab
from torch.optim import lr_scheduler
import flgo.algorithm.fedavg as fedavg
import flgo.utils.fmodule as fuf
from torch.utils.data import Dataset
from torchvision.transforms import RandomCrop, RandomHorizontalFlip
import torchvision
import torch
import utils.submodule as usub
from utils.pdmodule import *
MIN_RATIO = 0.0625
class Server(fab.BasicServer):
    def initialize(self, *args, **kwargs):
        self.init_local_model()

    def init_local_model(self):
        def get_model_size(model):
            return sum(p.numel() for p in model.parameters()) * 4 / (1024 ** 2)
        total_size = get_model_size(self.model)
        all_p = np.arange(MIN_RATIO, 1.001, MIN_RATIO)
        size_ratio = [1.0*get_model_size(self.model.__class__(p))/total_size for p in all_p]
        for c in self.clients:
            pidx = 0
            while pidx+1 < len(all_p) and (size_ratio[pidx+1] <= c._capacity): pidx += 1
            c.p = all_p[pidx]
            c.model = self.model.__class__(c.p)

    def run(self):
        tune_key, tune_direction = self.gv.logger.get_es_key(), self.gv.logger.get_es_direction()
        tune_key = tune_key.split('_')[-1]
        for c in self.clients: c.tune_key, c.tune_direction = tune_key, tune_direction
        self.gv.logger.time_start('Total Time Cost')
        # evaluating initial model performance
        self.gv.logger.info("--------------Initial Evaluation--------------")
        self.gv.logger.time_start('Eval Time Cost')
        self.gv.logger.log_once()
        self.gv.logger.time_end('Eval Time Cost')
        # Standalone Training
        for c in self.clients: c.finetune()
        self.gv.logger.info("=================End==================")
        self.gv.logger.time_end('Total Time Cost')
        self.gv.logger.time_start('Eval Time Cost')
        self.gv.logger.log_once()
        self.gv.logger.time_end('Eval Time Cost')
        # save results as .json file
        self.gv.logger.save_output_as_json()
        return

class Client(fab.BasicClient):
    def finetune(self):
        dataloader = self.calculator.get_dataloader(self.train_data, batch_size=self.batch_size)
        self.model.to(self.device)
        optimizer = self.calculator.get_optimizer(self.model, lr=self.learning_rate, weight_decay=self.weight_decay, momentum=self.momentum)
        epoch_iter = trange(self.num_epochs+1)
        op_model_dict = copy.deepcopy(self.model.state_dict())
        op_met = self.test(self.model, 'val')
        # op_met = {'_'.join(['local', 'val', k] if self.tune_key.startswith('local') else ['val', k]): v for k,v in op_met.items()}
        op_epoch = 0
        for e in epoch_iter:
            if e<self.num_epochs:
                for batch_id, batch_data in enumerate(dataloader):
                    optimizer.zero_grad()
                    batch_data = self.calculator.to_device(batch_data)
                    loss = self.calculator.compute_loss(self.model, batch_data)['loss']
                    loss.backward()
                    if self.clip_grad > 0: torch.nn.utils.clip_grad_norm_(parameters=self.model.parameters(), max_norm=self.clip_grad)
                    optimizer.step()
                val_metric = self.test(self.model, 'val')
                # val_metric = {'_'.join(['local', 'val', k] if self.tune_key.startswith('local') else ['val', k]): v for k, v in val_metric.items()}
                if val_metric[self.tune_key]*self.tune_direction>op_met[self.tune_key]:
                    op_met = val_metric
                    op_epoch = e+1
                    op_model_dict = copy.deepcopy(self.model.state_dict())
                else:
                    if self.option['early_stop']>0 and (e+1 - op_epoch>=self.option['early_stop']):
                        epoch_iter.set_description("\t".join(["Client {}".format(self.id), "Optimal Epoch {}/{}".format(op_epoch, self.num_epochs)]+['{}: {:.4f}'.format(k, v) for k, v in op_met.items()]))
                        self.model.load_state_dict(op_model_dict)
                        break
                epoch_iter.set_description("\t".join(["Client {}".format(self.id), "Epoch {}/{}".format(e+1, self.num_epochs)] + ['{}: {:.4f}'.format(k,v) for k,v in val_metric.items()]))
            else:
                epoch_iter.set_description("\t".join(["Client {}".format(self.id), "Optimal Epoch {}/{}".format(op_epoch, self.num_epochs)] + ['{}: {:.4f}'.format(k, v) for k, v in op_met.items()]))
                self.model.load_state_dict(op_model_dict)
        return



"""
This module is designed for creating challenging federated scenarios like noisy features, poisoned data, and truncated dataset based on existing tasks.
Each decorator can be applied to an existing runner by directly invoking the decorator and passing the runner as an argument. For example,
>>> runner = flgo.init(task, fedavg, option) # create runner
>>> import flgo.decorator as fd
>>> decorator = fd.TrainingDataReducer(0.8)
>>> decorator(runner)
>>> runner.run()
"""
import os.path
import shutil
import numpy as np
from toolz import isiterable
import torch.utils.data as tud

class BasicDecorator:
    """
    Basic decorator. Each customized decorator should inherent from this class and overwrite the __call__ method.
    Remark: it's recommended to call self.register_runner at the end of each __call__ method. This will create an
    additional directory to save the records of the decorated task, preventing the original records with the same
    name to be overwritten.
    """
    def register_runner(self, runner):
        # redirect task path for outputs
        original_task = runner.gv.logger.option['task']
        new_task = os.path.join(original_task, str(self))
        if not os.path.exists(new_task):
            os.makedirs(os.path.join(new_task, 'record'))
            shutil.copy(os.path.join(original_task, 'info'), new_task)
        runner.option['task'] = runner.option['task'] + str(self)
        runner.gv.logger.option['task'] = new_task
        runner.gv.logger.task_path = new_task

class ClientRemover(BasicDecorator):
    """
    Exclude partial clients from the training process

    Args:
        remove_idxs (list): the indices of clients to be excluded
    """
    def __init__(self, remove_idx=[]):
        if not isiterable(remove_idx): remove_idx=[remove_idx]
        self.remove_idx = remove_idx
        self.removed_clients = []
        self.preserved_clients = None

    def __call__(self, runner, *args, **kwargs):
        clients = runner.clients
        if self.remove_idx!=[]:
            assert max(self.remove_idx) < len(clients)
            self.removed_clients = [runner.clients[eid] for eid in self.remove_idx]
            self.preserved_clients = [runner.clients[cid] for cid in range(len(runner.clients)) if cid not in self.remove_idx]
            runner.reset_clients(self.preserved_clients)
        else:
            self.removed_clients = []
            self.preserved_clients = runner.clients
        self.register_runner(runner)
        return

    def __str__(self):
        return f"ClientRemover{len(self.removed_clients)}"

class RandomClientRemover(BasicDecorator):
    """
    Randomly exclude some clients from training process

    Args:
        remove_ratio (float): the ratio of clients to be excluded
    """
    def __init__(self, remove_ratio:float=0.1):
        self.remove_ratio = max(min(remove_ratio, 1.0), 0.0)
        self.removed_clients = []
        self.preserved_clients = None

    def __call__(self, runner, *args, **kwargs):
        num_clients = len(runner.clients)
        remove_idxs = np.random.choice(list(range(num_clients)), size=self.remove_ratio, replace=False).tolist()
        self.removed_clients = [runner.clients[eid] for eid in remove_idxs]
        self.preserved_clients = [runner.clients[cid] for cid in range(len(runner.clients)) if cid not in remove_idxs]
        self.register_runner(runner)

    def __str__(self):
        return f"RandomClientRemover{self.remove_ratio}"

class TrainingDataReducer(BasicDecorator):
    """
    Reduce training data at each client side

    Args:
        preserve_ratio (float): the ratio of training data to be preserved
    """
    def __init__(self, preserve_ratio=0.5):
        self.preserve_ratio = preserve_ratio

    def __call__(self, runner, *args, **kwargs):
        for c in runner.clients:
            new_train_data = tud.Subset(c.train_data, np.random.choice(list(range(len(c.train_data))), size=max(1, int(len(c.train_data) * self.preserve_ratio)), replace=False).tolist())
            c.set_data(new_train_data, 'train')
        self.register_runner(runner)

    def __str__(self):
        return f"TrainingDataReducer{self.preserve_ratio}"

class TrainingDataTruncation(BasicDecorator):
    """
    Truncating training data at each client side

    Args:
        max_size (int): the maximum size of training data
    """
    def __init__(self, max_size: int=1000):
        self.max_size = max_size

    def __call__(self, runner, *args, **kwargs):
        for c in runner.clients:
            if len(c.train_data)>self.max_size:
                new_train_data = tud.Subset(c.train_data, np.random.choice(list(range(len(c.train_data))), size=self.max_size, replace=False).tolist())
                c.set_data(new_train_data, 'train')
        self.register_runner(runner)

    def __str__(self):
        return f"TrainingDataTruncation{self.max_size}"

class SequentialDecorator(BasicDecorator):
    """
    Combine several decorators into a single decorator

    Args:
        decorators (list): the list of decorators
    """
    def __init__(self, decorators):
        self.decorators = decorators

    def __call__(self, runner):
        for dec in self.decorators:
            dec(runner)
        self.register_runner(runner)

    def __str__(self):
        return f"Seq_{'-'.join([str(d) for d in self.decorators])}"




import numpy as np
import torch.utils.data as tud
import flgo.decorator as fd

class LabelFlipper(fd.BasicDecorator):
    """
    Flip labels in datasets.

    Args:
        ratio_flip (float): the ratio of clients that will flip their labels
        label_map (dict): the label flipping map, e.g. {1:0} refers to flipping label 1 to label 0
    """
    class YFlipDataset(tud.Dataset):
        def __init__(self, data, lb_map):
            self.data = data
            self.lb_map = lb_map

        def __getitem__(self, index):
            x,y = self.data[index]
            new_y = self.lb_map[y] if y in self.lb_map else y
            return x,new_y

        def __len__(self):
            return len(self.data)

    def __init__(self, ratio_flip:float, label_map:dict={}):
        self.ratio_flip = ratio_flip
        self.label_map = label_map
        self.flipped_clients = []

    def __call__(self, runner, *args, **kwargs):
        num_flip = int(len(runner.clients) * self.ratio_flip)
        if self.ratio_flip > 0.: num_flip = max(1, num_flip)
        self.flipped_clients = np.random.choice(list(range(len(runner.clients))), num_flip, replace=False).tolist()
        for cid in self.flipped_clients:
            c = runner.clients[cid]
            flipped_train_data = self.flip_data(c.train_data)
            c.set_data(flipped_train_data, 'train')
        self.register_runner(runner)

    def flip_data(self, train_data):
        return self.YFlipDataset(train_data, self.label_map)

    def __str__(self):
        return f"LabelFlipper{self.ratio_flip}_F{len(self.label_map)}"






import torch
import torch.utils.data as tud
import flgo.decorator as fd
class GaussianNoiser(fd.BasicDecorator):
    """
    This noiser is a non-official implementation of Gaussian noise in 'Federated Learning on Non-IID Data Silos: An Experimental Study.'.
    The official implementation is at https://github.com/Xtra-Computing/NIID-Bench.

    Args:
        sigma (float): Standard deviation of the Gaussian noise.
        dynamic (bool): whether to use dynamic noise or not.
    """
    class NoiseDataset(tud.Dataset):
        def __init__(self, data, sigma, dynamic=True):
            self.data = data
            self.sigma = sigma
            self.dynamic = dynamic
            if not self.dynamic:
                x = data[0][0]
                self.noise = [torch.randn_like(x)*self.sigma for _ in range(len(data))]
            else:
                self.noise = None

        def __getitem__(self, index):
            x,y = self.data[index]
            if self.dynamic:
                new_x = x + torch.randn_like(x)*self.sigma
            else:
                new_x = x + self.noise[index]
            return new_x,y

        def __len__(self):
            return len(self.data)

    def __init__(self, sigma:float, dynamic=True):
        self.sigma = sigma
        self.dynamic = dynamic

    def __call__(self, runner, *args, **kwargs):
        sigmas = [self.sigma*(i+1.)/len(runner.clients) for i in range(len(runner.clients))]
        for c,sc in zip(runner.clients, sigmas):
            c.set_data(self.NoiseDataset(c.train_data, sc, self.dynamic), 'train')
            if c.val_data is not None and len(c.val_data) > 0:
                c.set_data(self.NoiseDataset(c.val_data, sc, self.dynamic), 'val')
            if c.test_data is not None and len(c.test_data) > 0:
                c.set_data(self.NoiseDataset(c.test_data, sc, self.dynamic), 'test')
        self.register_runner(runner)

    def __str__(self):
        return f"Gaussian-0-{self.sigma}*iN"

class HardGaussianNoiser(fd.BasicDecorator):
    """
    This noiser adds noise to features with different means and standard deviations across clients.

    Args:
        std_mu (float): Standard deviation of the clients' local means of Gaussian noise.
        std_sigma (float): Standard deviation of the Gaussian noise.
        dynamic (bool): whether to use dynamic noise or not.
    """
    class NoiseDataset(tud.Dataset):
        def __init__(self, data, mu, sigma, dynamic=True):
            self.data = data
            self.sigma = sigma
            self.mu = mu
            self.dynamic = dynamic
            if not self.dynamic:
                x = data[0][0]
                self.noise = [mu+torch.randn_like(x)*self.sigma for _ in range(len(data))]
            else:
                self.noise = None

        def __getitem__(self, index):
            x, y = self.data[index]
            if self.dynamic:
                new_x = x + self.mu + torch.randn_like(x) * self.sigma
            else:
                new_x = x + self.noise[index]
            return new_x, y

        def __len__(self):
            return len(self.data)

    def __init__(self, std_mu:float, std_sigma: float, dynamic=True):
        self.std_mu = std_mu
        self.std_sigma = std_sigma
        self.dynamic = dynamic

    def __call__(self, runner, *args, **kwargs):
        example = runner.clients[0].train_data[0][0]
        for c in runner.clients:
            cmu = torch.randn_like(example)*self.std_mu
            csigma = torch.randn(1)*self.std_sigma
            c.set_data(self.NoiseDataset(c.train_data, cmu, csigma, self.dynamic), 'train')
            if c.val_data is not None and len(c.val_data) > 0:
                c.set_data(self.NoiseDataset(c.val_data, cmu, csigma, self.dynamic), 'val')
            if c.test_data is not None and len(c.test_data) > 0:
                c.set_data(self.NoiseDataset(c.test_data, cmu, csigma, self.dynamic), 'test')
        self.register_runner(runner)

    def __str__(self):
        return f"Gaussian_Nmu-0-{self.std_mu}_Nsigma-0-{self.std_sigma}"
